/*
 *	Simple Linux Clustering Extensions -- Master Daemon
 *
 *	Copyright (c) 1999 Martin Mares <mj@ucw.cz>
 *		  (c) 1999 Michal Jezek <mjez6230@ss1000.ms.mff.cuni.cz>
 *		  (c) 1999 Jan Lana <jlan3028@ss1000.ms.mff.cuni.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "lib.h"
#include "sched.h"

struct node_config nodes[SLICE_MAX_NODES];
int num_nodes;

static void
dump_status(int which)
{
#ifdef DEBUG_SCHED
  int lo, hi, i;
  char *states[] = { "DOWN", "FULL", "OK" };

  if (which < 0)
    lo = 0, hi = num_nodes - 1;
  else
    lo = hi = which;
  for(i=lo; i<=hi; i++)
    if (nodes[i].addr.s_addr)
      {
	struct node_config *c = &nodes[i];
	struct slice_status *s = &c->status;
	struct sched_status *ss = &c->sched;
	printf("%2d: %-15s  M=%6d/%6d S=%6d/%6d L=%2d.%02d S=%2d.%02d D=%2d.%02d R=%2d P=%2d.%02d %s\n",
	       i,
	       inet_ntoa(c->addr),
	       s->freemem, s->totalmem,
	       s->freeswap, s->totalswap,
	       s->load / 100, s->load % 100,
	       ss->s_prior / (1<<FSHIFT), ss->s_prior % (1<<FSHIFT) / 10,
	       ss->d_prior / (1<<FSHIFT), ss->d_prior % (1<<FSHIFT) / 10,
	       ss->runs,
	       ss->priority / (1<<FSHIFT), ss->priority % (1<<FSHIFT) / 10,
	       states[s->admit]);
      }
#endif
}

static void
load_node_list(char *fn)
{
  FILE *f;
  char buf[256];
  int node = 0;
  struct hostent *he;

  if (!(f = fopen(fn, "r")))
    die("Unable to open %s: %m", fn);
  while (fgets(buf, sizeof(buf)-1, f))
    {
      char *c = strchr(buf, '\n');
      if (*c)
	*c = 0;
      he = gethostbyname(buf);
      if (!he)
	die("Unable to resolve %s: %m", buf);
      if (node >= SLICE_MAX_NODES)
	die("Too many nodes, ask a wizard to enlarge me");
      memcpy(&nodes[node].addr, he->h_addr, sizeof(struct in_addr));
      node++;
    }
  fclose(f);
  if (!node)
    die("No nodes defined");
  num_nodes = node;
}

static void
feed_kernel(void)
{
  int i;
 
   for(i=0; i<num_nodes; i++)
     define_node(i, &nodes[i].addr, !i);
}

static void
handle_init(struct slicemsg *m)
{
  ALLOCA_MSG_BYTES(rm, struct in_addr, r, num_nodes * sizeof(struct in_addr));
  int i;

  if (!m->from)
    {
      puts("Somebody is claiming he's the master!");
      return;
    }
  printf("Node %d is booting.\n", m->from);
  mk_reply(rm, m);
  for(i=0; i<num_nodes; i++)
    r[i] = nodes[i].addr;
  sm_send(rm);
}

static void
handle_preexec(struct slicemsg *m)
{
  ALLOCA_MSG(rm, r, sm_empty);

  DBG("Process placement request from node %d\n", m->from);
  mk_reply(rm, m);
  rm->arg = sched_get_node(m->from);
  DBG("-> %d\n", rm->arg);
  sm_send(rm);
  dump_status(rm->arg);
}

int
handle_pkt(struct slicemsg *m)
{
  if (m->from >= num_nodes)
    {
      printf("Received packet with invalid source %d\n", m->from);
      return 1;
    }
  switch (m->cmd)
    {
    case SLICE_CMD_INIT:
      handle_init(m);
      break;
    case SLICE_CMD_NODE_STATUS:
      DBG("Received node status from %d\n", m->from);
      memcpy(&nodes[m->from].status, m->data, sizeof(struct slice_status));
      sched_change_stat(m->from);
      dump_status(m->from);
      break;
    case SLICE_CMD_DEATH_NOTIFY:
      printf("Node %d is dead.\n", m->from);
      bzero(&nodes[m->from].status, sizeof(struct slice_status));
      sched_death(m->from);
      dump_status(m->from);
      break;
    case SLICE_CMD_PREEXEC:
      handle_preexec(m);
      break;
    default:
      return 0;
    }
  return 1;
}

void
handle_timer(void)
{
  static int ticks = 1;
  
  if (!--ticks)
    {
      ALLOCA_MSG(m, s, slice_status);

      mk_status(m);
      handle_pkt(m);
      ticks = STATUS_REPORT_PERIOD;
  }
  sched_tick();
  dump_status(-1);
}

int
main(int argc, char **argv)
{
  if (argc > 1 && !strcmp(argv[1], "-d"))
    {
      argc--, argv++;
      debug = 1;
    }
  if (argc != 2)
    die("Usage: cmasterd [-d] <node-list>");
  load_node_list(argv[1]);
  timer_period = 1;
  daemon_init();
  feed_kernel();
  main_loop();
  return 0;
}
