/*
 *	Simple Linux Clustering Extensions -- Daemon Skeleton
 *
 *	Copyright (c) 1999 Martin Mares <mj@ucw.cz>
 *		  (c) 1999 Michal Jezek <mjez6230@ss1000.ms.mff.cuni.cz>
 *		  (c) 1999 Jan Lana <jlan3028@ss1000.ms.mff.cuni.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>

#include "lib.h"

int this_node;
int debug;
int timer_period;

static void
int_handler(int signum)
{
  puts("Shutting down cluster connections");
  slicectl(SLICE_OP_CLOSE_ALL, NULL);
  exit(1);
}

static void
alarm_handler(int signum)
{
  DBG("Tick tock, here's the clock...\n");
  handle_timer();
  alarm(timer_period);
}

void
main_loop(void)
{
  const int maxlen = 1024;
  struct slicemsg *m = alloca(sizeof(struct slicemsg) + maxlen);

  alarm_handler(0);
  for(;;)
    {
      int x;
      m->length = maxlen;
      x = slicectl(SLICE_OP_GET_PACKET, m);
      if (x < 0)
	die("get_packet: %m");
      if (debug)
	pkt_dump(m);
      if (m->to != this_node)
	{
	  printf("Received packet with invalid destination %d\n", m->to);
	  continue;
	}
      else if (!handle_pkt(m))
	{
	  printf("Received unsupported packet type %04x from node %d. Pity, isn't it?\n", m->cmd, m->from);
	  if ((m->cmd & SLICE_CMD_TYPE_MASK) == SLICE_CMD_TYPE_REQUEST)
	    {
	      ALLOCA_MSG(rm, r, sm_empty);
	      mk_reply(rm, m);
	      rm->arg = -EINVAL;
	      sm_send(rm);
	    }
	}
    }
}

void
daemon_init(void)
{
  struct sigaction sa;

  slicectl(SLICE_OP_DEBUG, (struct slicemsg *) debug);
  slicectl(SLICE_OP_CLOSE_ALL, NULL);
  bzero(&sa, sizeof(sa));
  sa.sa_handler = int_handler;
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGTERM, &sa, NULL);
  sa.sa_handler = alarm_handler;
  sa.sa_flags = SA_RESTART;
  sigaction(SIGALRM, &sa, NULL);
}

void
define_node(int n, struct in_addr *a, int this)
{
  ALLOCA_MSG(m, r, in_addr);

  m->to = n;
  r[0] = *a;
  if (slicectl(this ? SLICE_OP_DEFINE_THIS_NODE : SLICE_OP_DEFINE_NODE, m))
    die("define_node: %m");
}

void
mk_reply(struct slicemsg *r, struct slicemsg *m)
{
  r->from = this_node;
  r->to = m->from;
  r->cmd = REPLY_TO(m->cmd);
  r->ident = m->ident;
}

void
mk_status(struct slicemsg *m)
{
  struct slice_status *s = (void *) &m->data;
  static FILE *meminfo, *loadavg;
  char buf[256], *c;
  int cached, buffers, lahi, lalo;

  m->from = this_node;
  m->cmd = SLICE_CMD_NODE_STATUS;

  if (!meminfo)
    {
      if (!(meminfo = fopen("/proc/meminfo", "r")))
	die("Unable to open /proc/meminfo: %m");
    }
  else
    {
      rewind(meminfo);
      fflush(meminfo);
    }
  s->freemem = -1;
  s->totalmem = -1;
  s->freeswap = -1;
  s->totalswap = -1;
  while (fgets(buf, sizeof(buf)-1, meminfo))
    {
      c = strchr(buf, '\n');
      if (!c)
	die("/proc/meminfo: Parse error");
      *c = 0;
      if (!strncmp(buf, "MemTotal:", 9))
	sscanf(buf+9, "%d", &s->totalmem);
      if (!strncmp(buf, "MemFree:", 8))
	sscanf(buf+8, "%d", &s->freemem);
      if (!strncmp(buf, "Cached:", 7))
	sscanf(buf+7, "%d", &cached);
      if (!strncmp(buf, "Buffers:", 8))
	sscanf(buf+8, "%d", &buffers);
      if (!strncmp(buf, "SwapTotal:", 10))
	sscanf(buf+10, "%d", &s->totalswap);
      if (!strncmp(buf, "SwapFree:", 9))
	sscanf(buf+9, "%d", &s->freeswap);
    }
  if (s->freemem < 0 || s->totalmem < 0 || s->freeswap < 0 || s->totalswap < 0 || cached < 0 || buffers < 0)
    die("/proc/meminfo: Incomplete information");
  s->freemem += cached + buffers;

  if (!loadavg)
    {
      if (!(loadavg = fopen("/proc/loadavg", "r")))
	die("Unable to open /proc/loadavg: %m");
    }
  else
    {
      rewind(loadavg);
      fflush(loadavg);
    }
  if (fscanf(loadavg, "%d.%d", &lahi, &lalo) != 2)
    die("/proc/loadavg: Parse error");
  s->load = 100*lahi + lalo;

  s->admit = 2;				/* FIXME: Use a better criterion */
}
